from .. import base

annotation_check_cut = base.LocalizedString(
    {
        'EN': """
<div>
    <p>
        You will see solutions to tasks completed by the instruction below. Evaluate solutions according to this instruction.
    </p>
</div>
<h1>Instructions according to which solutions were given</h1>
""",
        'RU': """
<div>
    <p>
        Вы увидите решения заданий, выполненные по инструкции ниже. Проверьте решения на соответствие инструкции.
    </p>
</div>
<h1>Инструкция, по которой выполнены решения</h1>
""",
        'TR': """
<div>
    <p>
        Aşağıdaki talimatlara göre tamamlanan iş çözümlerini göreceksiniz. Kontrol çözümleri uyum için talimatlar.
    </p>
</div>
<h1>Çözümlerin yürütüldüğü talimat</h1>
""",
        'KK': """
<div>
    <p>
        Төмендегі нұсқауларға сәйкес орындалған тапсырмалардың шешімдерін көресіз. Шешімдердің нұсқаулыққа сәйкестігін тексеріңіз.
    </p>
</div>
<h1>Шешімдер орындалған Нұсқаулық</h1>
""",
    }
)
